// -*-c++-*-

#include "osgGLUT/Viewer"
#include "osgDB/ReadFile"
#include "osg/Vec3"

#include "osg/Geode"
#include "osg/GeoSet"
#include "osg/StateSet"
#include "osg/Material"
#include "osg/Texture"
#include "osg/TexMat"
#include "osg/CullFace"

#include <osgGA/TrackballManipulator>

#include <GL/glut.h>
#include <math.h>

#include "MorphVec3.h"
#include "MpegImageStream.h"


/*
 * Create morphed textured geometry
 */
osg::Geode* morphGeom(osg::Vec3* coords,
                      osg::Vec3* normals,
                      osg::Vec2* texCoords,
                      osg::Image* image,
                      osg::TexMat* texMat)
{
    /*
     * GeoSet
     */
    osg::GeoSet* gset = new osg::GeoSet();

    gset->setPrimType(osg::GeoSet::POLYGON);
    gset->setNumPrims(1);

    int* primLen = new int[1];
    primLen[0] = 4;
    gset->setPrimLengths(primLen);

    gset->setCoords(coords);

    gset->setNormals(normals);
    gset->setNormalBinding(osg::GeoSet::BIND_PERPRIM);

    gset->setTextureCoords(texCoords);
    gset->setTextureBinding(osg::GeoSet::BIND_PERVERTEX);

    /*
     * StateSet
     */
    osg::StateSet* state = new osg::StateSet;

    osg::Material* mtl = new osg::Material();
    osg::Vec4 white( 1.0f, 1.0f, 1.0f, 1.0f );
    mtl->setEmission( osg::Material::FRONT_AND_BACK, white );
    mtl->setAmbient( osg::Material::FRONT_AND_BACK, white );
    mtl->setDiffuse( osg::Material::FRONT_AND_BACK, white );
    mtl->setSpecular( osg::Material::FRONT_AND_BACK, white );
    state->setAttribute(mtl);

    osg::Texture* tex = new osg::Texture;
    if (!image) {
        image = osgDB::readImageFile("lz.rgb");
    }
    tex->setImage(image);
    tex->setFilter(osg::Texture::MIN_FILTER, osg::Texture::NEAREST);
    tex->setWrap(osg::Texture::WRAP_S, osg::Texture::CLAMP);
    tex->setWrap(osg::Texture::WRAP_T, osg::Texture::CLAMP);
    tex->setSubloadMode(osg::Texture::IF_DIRTY);
    state->setTextureAttributeAndModes(0, tex, osg::StateAttribute::ON);

    if (texMat)
        state->setTextureAttributeAndModes(0, texMat, osg::StateAttribute::ON);

    // don't cull faces
    osg::CullFace* cull = new osg::CullFace;
    state->setAttributeAndModes(cull, osg::StateAttribute::OFF);

    /*
     * Geode
     */
    osg::Geode* geode = new osg::Geode;
    geode->setStateSet( state );
    geode->addDrawable( gset );

    return geode;
}


/*
 * Morph callback
 */
void morphFunc(int value)
{
    MorphVec3* morph = (MorphVec3*) value;

    static float t = 0.0f;
    static float tdir = 0.01f;
    morph->update(t);

    t += tdir;
    if (t < 0.0f) {
        t = 0.0f;
        tdir = -tdir;
    }
    else if (t > 1.0f) {
        t = 1.0f;
        tdir = -tdir;
    }

    glutTimerFunc(75, morphFunc, value);
}


/*
 * Main
 */
int main(int argc, char** argv)
{
    // coordinates
    osg::Vec3 coord0[4];
    coord0[0].set( -1.0f, 0.0f, -1.0f );
    coord0[1].set(  1.0f, 0.0f, -1.0f );
    coord0[2].set(  1.0f, 0.0f,  1.0f );
    coord0[3].set( -1.0f, 0.0f,  1.0f );

    osg::Vec3 coord1[4];
    coord1[0].set( -2.0f, 0.0f, -1.0f );
    coord1[1].set(  2.0f, 0.0f, -1.0f );
    coord1[2].set(  2.0f, 0.0f,  1.0f );
    coord1[3].set( -2.0f, 0.0f,  1.0f );
    int ncoord = sizeof(coord0) / sizeof(coord0[0]);

    // normals
    osg::Vec3 normals[1];
    normals[0].set( 0.0f, 1.0f, 0.0f );

    // texture coordinates
    osg::Vec2 texCoords[4];
    texCoords[0].set(0.0f, 0.0f);
    texCoords[1].set(1.0f, 0.0f);
    texCoords[2].set(1.0f, 1.0f);
    texCoords[3].set(0.0f, 1.0f);

    // create coordinate morph
    MorphVec3 coordMorph;
    coordMorph.setCoords(ncoord, coord0, coord1);

    // open MpegImageStream
    osg::MpegImageStream* mpeg = NULL;
    osg::TexMat* texMat = NULL;
    if (argc > 1) {
        mpeg = new osg::MpegImageStream(argv[1]);
        mpeg->start();
        texMat = (osg::TexMat*) mpeg->getTexMat();
    }

    // Create morphed geometry
    osg::Geode* geode = morphGeom((osg::Vec3*) coordMorph.getMorphCoords(),
                                  normals, texCoords, mpeg, texMat);
    coordMorph.addGeode(geode);

    // Setup GLUT
    glutInit(&argc, argv);
    //glutTimerFunc(100, morphFunc, (int) &coordMorph);

    osgGLUT::Viewer viewer;
    viewer.addViewport( geode );
    viewer.registerCameraManipulator(new osgGA::TrackballManipulator);
    viewer.open();

    viewer.run();

    return 0;
}
